from typing import Callable, NamedTuple, Tuple, Union

import jax.numpy as jnp
from jax import jacrev, jit, lax, ops, random, tree_map, vmap
from jax.scipy.stats import norm
from functools import partial

from bayex.gp import DataTypes, GParameters, predict, round_vars, train
from bayex.types import Array


class OptimizerParameters(NamedTuple):
    """
    Object holding the results of the optimization.
    """

    target: Union[Array, float]
    params: Array
    f: Callable
    params_all: Array
    target_all: Array


def jacobian(f: Callable) -> Callable:
    return jit(jacrev(f))


def expected_improvement(
    x_pred: Array,
    params: GParameters,
    x: Array,
    y: Array,
    xi: float = 0.01,
    dtypes: Union[dict, None] = None,
) -> Array:
    """
    Computes the expected improvement at points x_pred over a
    Gaussian process trained on x and y.

    Parameters:
    -----------
    x_pred: The points at which the improvement is computed.
    params: Trained hyperparameters of the GP.
    x: Sampled points.
    y: Target values of the sampled points.
    xi: Parameter to balance exploration-exploitation.

    Returns:
    --------
    ei: The expected improvement of x_pred on the trained GP.
    """
    y_max = y.max()
    mu, std = predict(params, x, y, dtypes, xt=x_pred)
    improvement = mu.T - y_max - xi
    z = improvement / (std + 1e-3)
    ei = improvement * norm.cdf(z) + std * norm.pdf(z)
    return ei


def replace_nan_values(arr: Array) -> Array:
    """
    Replaces the NaN values (if any) in arr with 0.

    Parameters:
    -----------
    arr: The array where NaN is removed from.

    Returns:
    --------
    The array with all the NaN elements replaced with 0.
    """
    # todo(alonfnt): Find a more robust solution.
    return jnp.where(jnp.isnan(arr), 0, arr)


@jit
def suggest_next(
    key: Array,
    params: GParameters,
    x: Array,
    y: Array,
    bounds: Array,
    dtypes: DataTypes,
    n_seed: int = 1000,
    lr: float = 0.1,
    n_epochs: int = 150,
) -> Tuple[Array, Array]:
    """
    Suggests the new point to sample by optimizing the acquisition function.

    Parameters:
    -----------
    key: The pseudo-random generator key used for jax random functions.
    params: Hyperparameters of the Gaussian Process Regressor.
    x: Sampled points.
    y: Sampled targets.
    bounds: Array of (2, dim) shape with the lower and upper bounds of the
            variables.y_max: The current maximum value of the target values Y.
    dtypes: The type of non-real variables in the target function.
    n_seed (optional): the number of points to probe and minimize until
            finding the one that maximizes the acquisition functions.
    lr (optional): The step size of the gradient descent.
    n_epochs (optional): The number of steps done on the descent to minimize
            the seeds.


    Returns:
    --------
    A tuple with the parameters that maximize the acquisition function and a
    jax PRGKey to be used in the next sampling.
    """

    key1, key2 = random.split(key, 2)
    dim = bounds.shape[0]

    domain = random.uniform(
        key1, shape=(n_seed, dim), minval=bounds[:, 0], maxval=bounds[:, 1]
    )

    acq = partial(expected_improvement, params=params, x=x, y=y, dtypes=dtypes)

    J = jacobian(lambda x: acq(x.reshape(-1, dim)).reshape())
    HS = vmap(lambda x: x + lr * J(x))

    domain = lax.fori_loop(0, n_epochs, lambda _, d: HS(d), domain)
    domain = jnp.clip(
        domain.reshape(-1, dim), a_min=bounds[:, 0], a_max=bounds[:, 1]
    )
    domain = replace_nan_values(domain)
    domain = round_vars(domain, dtypes.integers)

    ys = acq(domain)
    next_X = domain[ys.argmax()]
    return next_X, key2


@partial(jit, static_argnums=(1, 2))
def _extend_array(arr: Array, pad_width: int, axis: int) -> Array:
    """
    Extends the array pad_width only on one direction and fills it with
    the last value of that axis.
    TODO: consider donate_argnums=0 if the device allows it.
    """
    pad_shape = [(0, 0)] * arr.ndim
    pad_shape[axis] = (0, pad_width)
    return jnp.pad(arr, pad_shape, mode="edge")


def optim(
    f: Callable,
    constrains: dict,
    seed: int = 42,
    n_init: int = 5,
    n: int = 10,
    xi: float = 0.01,
    ctypes: dict = None,
) -> OptimizerParameters:
    """
    Finds the inputs of 'f' that yield the maximum value between the given
    'constrains', after 'n_init' + 'n' iterations.

    Parameters:
    -----------
    f: Function to optimize.
    constrains: Dictionary with the domain of each input variable.
    seed: Pseudo-random number generator seed for reproducibility
    n_init: Number of initial evaluations before suggesting optimized samples.
    n: Number of sampling iterations.
    xi: Parameter to balance exploration-exploitation.
    ctypes: The type of non-real variables in the target function.

    Returns:
    --------
    The parameters that maximize the given 'f'.
    """

    assert n > 0, "Num of iterations n should be a positive integer"

    key = random.PRNGKey(seed)
    dim = len(constrains)
    _vars = f.__code__.co_varnames[: f.__code__.co_argcount]
    _sorted_constrains = {k: constrains[k] for k in _vars}

    if ctypes is not None:
        _sorted_types = {k: ctypes[k] for k in _vars if k in ctypes}
        dtypes = DataTypes(
            integers=[
                _vars.index(k) for k, v in _sorted_types.items() if v == int
            ]
        )
    else:
        dtypes = DataTypes(integers=[])

    bounds = jnp.asarray(list(_sorted_constrains.values()))

    X = random.uniform(
        key,
        shape=(n_init, dim),
        minval=bounds[:, 0],
        maxval=bounds[:, 1],
    )
    X = round_vars(X, dtypes.integers)
    Y = vmap(f)(*X.T)

    # Expand the array with the same last values to not perjudicate the gp.
    # the reason to apply it as a function is to avoid having twice the memory
    # usage, since JAX does not do inplace updates except after being
    # compiled.
    X = _extend_array(X, n, 0)
    Y = _extend_array(Y, n, 0)

    # Initialize the GP parameters
    params = GParameters(
        noise=jnp.zeros((1, 1)) - 5.0,
        amplitude=jnp.zeros((1, 1)),
        lengthscale=jnp.zeros((1, dim)),
    )
    momentums = tree_map(lambda x: x * 0, params)
    scales = tree_map(lambda x: x * 0 + 1, params)

    # todo(alonfnt): make it JAX friendly
    for i in range(n):
        idx = n_init + i
        params, momentums, scales = train(
            X, Y, params, momentums, scales, dtypes
        )
        max_params, key = suggest_next(key, params, X, Y, bounds, dtypes)
        X = ops.index_update(X, ops.index[idx, ...], max_params)
        Y = ops.index_update(Y, ops.index[idx], f(*max_params))

    best_target = float(Y.max())
    best_params = {k: v for (k, v) in zip(constrains.keys(), X[Y.argmax()])}
    optimizer_params = OptimizerParameters(
        target=best_target, params=best_params, f=f, params_all=X, target_all=Y
    )
    return optimizer_params



# Based on the Gaussian regression example in
# https://github.com/google/jax/blob/main/examples/gaussian_process_regression.py
from collections import namedtuple
from functools import partial
from typing import Any, Callable, Tuple

import jax.numpy as jnp
import jax.scipy as scipy
from jax import grad, jit, lax, ops, tree_map, tree_multimap, vmap

from bayex.types import Array

GParameters = namedtuple("GParameters", ["noise", "amplitude", "lengthscale"])
DataTypes = namedtuple("DataTypes", ["integers"])


def cov_map(cov_func: Callable, xs: Array, xs2: Array = None) -> Array:
    """
    Computes the covariance matrix of the given function with the
    given points.
    """
    if xs2 is None:
        return vmap(lambda x: vmap(lambda y: cov_func(x, y))(xs))(xs)
    else:
        return vmap(lambda x: vmap(lambda y: cov_func(x, y))(xs))(xs2).T


def softplus(x: Array) -> Array:
    return jnp.logaddexp(x, 0.0)


def exp_quadratic(x1: Array, x2: Array, ls: Array) -> Array:
    return jnp.exp(-jnp.sum((x1 - x2) ** 2 / ls ** 2))


def round_vars(arr: Array, indexes: list) -> Array:
    """
    The input variables corresponding to an integer-valued input variable are
    rounded to the closest integer value.
    """
    for idx in indexes:
        arr = ops.index_update(arr, ops.index[:, idx], jnp.round(arr[:, idx]))
    return arr


def gp(
    params: GParameters,
    x: Array,
    y: Array,
    dtypes: DataTypes,
    xt: Array = None,
    compute_ml: bool = False,
) -> Any:
    """
    Gaussian Processor Main function.
    It is used as the based for the trainig of the GP, as well
    as the computing of marginal likelihood and prediction.

    Parameters:
    -----------
    params: Hyperparameters of the GP.
    x: Sampled points.
    y: Target values of the sampled points.
    xt: Points on which to predict.
    compute_ml: Flag indicating whether return the marginal
                distribution or continue with the predictions.

    Returns:
    --------
    The mean and standard deviations of the resulting
    Gaussian Distributions.
    """
    n = x.shape[0]
    x = round_vars(x, dtypes.integers)
    noise, amp, ls = tree_map(softplus, params)
    kernel = partial(exp_quadratic, ls=ls)

    ymean = jnp.mean(y)
    y = y - ymean
    train_cov = amp * cov_map(kernel, x) + jnp.eye(n) * (noise + 1e-6)
    chol = scipy.linalg.cholesky(train_cov, lower=True)
    kinvy = scipy.linalg.solve_triangular(
        chol.T, scipy.linalg.solve_triangular(chol, y, lower=True)
    )
    if compute_ml:
        log2pi = jnp.log(2.0 * jnp.pi)
        ml = jnp.sum(
            -0.5 * jnp.dot(y.T, kinvy)
            - jnp.sum(jnp.log(jnp.diag(chol)))
            - (n / 2.0) * log2pi
        )
        ml -= jnp.sum(-0.5 * jnp.log(2 * 3.1415) - jnp.log(amp) ** 2)
        return -ml

    if xt is not None:
        xt = round_vars(xt, dtypes.integers)

    cross_cov = amp * cov_map(kernel, x, xt)
    mu = jnp.dot(cross_cov.T, kinvy) + ymean
    v = scipy.linalg.solve_triangular(chol, cross_cov, lower=True)
    var = amp * cov_map(kernel, xt) - jnp.dot(v.T, v)

    if n > 1:
        var = jnp.diag(var)

    std = jnp.sqrt(var)
    return mu, std


marginal_likelihood = partial(gp, compute_ml=True)
predict = jit(partial(gp, compute_ml=False))
grad_fun = jit(grad(marginal_likelihood))


@jit
def train(
    x: Array,
    y: Array,
    params: GParameters,
    momentums: GParameters,
    scales: GParameters,
    dtypes: DataTypes,
    lr: float = 0.01,
    nsteps: int = 20,
) -> Tuple[GParameters, GParameters, GParameters]:
    """
    Training function of the Gaussian Process Regressor.

    Parameters:
    -----------
    x: Sampled points.
    y: Target values of the sampled points.
    params: Hyperparameters of the GP.
    lr: Learning rate of the train step.
    nsteps: Number of epochs to train.

    Returns:
    --------
    Tuple with the trained `params`, `momentums` and `scales`.
    """

    def train_step(
        params: GParameters, momentums: GParameters, scales: GParameters
    ) -> Tuple:
        grads = grad_fun(params, x, y, dtypes=dtypes)
        momentums = tree_multimap(
            lambda m, g: 0.9 * m + 0.1 * g, momentums, grads
        )
        scales = tree_multimap(
            lambda s, g: 0.9 * s + 0.1 * g ** 2, scales, grads
        )
        params = tree_multimap(
            lambda p, m, s: p - lr * m / jnp.sqrt(s + 1e-5),
            params,
            momentums,
            scales,
        )
        return params, momentums, scales

    params, momentums, scales = lax.fori_loop(
        0,
        nsteps,
        lambda _, v: train_step(*v),
        (params, momentums, scales),
    )

    return params, momentums, scales



from enum import Enum
from typing import Any, Callable

import jax.numpy as jnp
from jax.tree_util import tree_map

from bayex.optim import OptimizerParameters
from bayex.types import Array


class Log(Enum):
    ALL = 1
    BEST = 2


def show_sampling(
    x: Array, y: Array, f: Callable, verbose: Log, min_width: int
) -> None:
    """
    Prints the sampled points during the optimization as well as the target
    value.

    Parameters:
    -----------
    x: The sampled points.
    y: The sampled targets.
    f: Target function.
    verbose: Level of detail in the shown information.
             - Log.ALL shows all the sampling points.
             - Log.BEST shows only the iterations that find better results.
    min_width: Minimum width of the column.
    """
    assert x.ndim == 2, "sampled points should be a 2d-array (throws, vars)"
    var_count = x.shape[1]
    assert var_count == f.__code__.co_argcount
    var_names = f.__code__.co_varnames[:var_count]
    max_width = max(tree_map(len, var_names))
    max_width = max_width if max_width > min_width else min_width
    col_template = "{:^" + str(max_width) + "}"
    row_template = col_template * (var_count + 2)
    print(row_template.format("#sample", "target", *var_names))
    col_template = "{:^" + str(max_width) + "g}"
    row_template = col_template * (var_count + 2)

    # I doubt whether is should be better to print in sorted mode
    # or is it better to show in the order they have been sampled
    y_prev = -jnp.inf
    for i, (xi, yi) in enumerate(zip(x, y)):
        if verbose == Log.BEST and yi < y_prev:
            continue
        print(row_template.format(i + 1, yi, *xi))
        y_prev = yi


def show_results(
    res: OptimizerParameters, verbose: Log = Log.ALL, min_width: int = 10
) -> None:
    """
    Prints the sampled points during the optimization as well as the target
    value.

    - This function is a wrapper for `show_samplings`.
    Parameters:
    -----------
    res: The object returned from the optimizer with the optimization
         parameters.
    verbose: Level of detail in the shown information.
             - Log.ALL shows all the sampling points.
             - Log.BEST shows only the iterations that find better results.
    min_width: Minimum width of the column.
    """
    show_sampling(
        x=res.params_all,
        y=res.target_all,
        f=res.f,
        verbose=verbose,
        min_width=min_width,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bayex']

package_data = \
{'': ['*']}

install_requires = \
['jax>=0.2.18,<0.3.0', 'jaxlib>=0.1.69,<0.2.0']

setup_kwargs = {
    'name': 'bayex',
    'version': '0.1.0',
    'description': 'Bayesian Optimization with Gaussian Processes powered by JAX',
    'long_description': "# BAYEX: Bayesian Optimization powered by JAX\n[![tests](https://github.com/alonfnt/bayex/actions/workflows/tests.yml/badge.svg)](https://github.com/alonfnt/bayex/actions/workflows/tests.yml)\n\nBayex is a high performance Bayesian global optimization library using Gaussian processes.\nIn contrast to existing Bayesian optimization libraries, Bayex is designed to use JAX as its backend.\n\nInstead of relaying on external libraries, Bayex only relies on JAX and its custom implementations, without requiring importing massive libraries such as `sklearn`.\n\n## What is Bayesian Optimization?\n\nBayesian Optimization (BO) methods are useful for optimizing functions that are expensive to evaluate, lack an analytical expression and whose evaluations can be contaminated by noise.\nThese methods rely on a probabilistic model of the objective function, typically a Gaussian process (GP), upon which an acquisition function is built.\nThe acquisition function guides the optimization process and measures the expected utility of performing an evaluation of the objective at a new point.\n\n## Why JAX?\nUsing JAX as a backend removes some of the limitations found on Python, as it gives us direct mapping to the XLA compiler.\n\nXLA compiles and runs the JAX code into several architectures such as CPU, GPU and TPU without hassle. But the device agnostic approach is not the reason to back XLA for future scientific programs. XLA provides with optimizations under the hood such as Just-In-Time compilation and automatic parallelization that make Python (with a NumPy-like approach) a suitable candidate on some High Performance Computing scenarios.\n\nAdditionally, JAX provides Python code with automatic differentiation, which helps identify the conditions that maximize the acquisition function.\n\n\n## Installation\nBayex can be installed using [PyPI](https://pypi.org/project/bayex/) via `pip`:\n```\npip install bayex\n```\nor from GitHub directly\n```\npip install git+git://github.com/alonfnt/bayex.git\n```\nFor more advance instructions please refer to the [installation guide](INSTALLATION.md).\n\n## Usage\nUsing Bayex is very straightforward:\n```python\nimport bayex\n\ndef f(x, y):\n    return -y ** 2 - (x - y) ** 2 + 3 * x / y - 2\n\nconstrains = {'x': (-10, 10), 'y': (0, 10)}\noptim_params = bayex.optim(f, constrains=constrains, seed=42, n=10)\n```\nshowing the results can be done with\n```python\n>> bayex.show_results(optim_params, min_len=13)\n   #sample      target          x            y\n      1        -9.84385      2.87875      3.22516\n      2        -307.513     -6.13013      8.86493\n      3        -19.2197      6.8417       1.9193\n      4        -43.6495     -3.09738      2.52383\n      5        -58.9488      2.63803      6.54768\n      6        -64.8658      4.5109       7.47569\n      7        -78.5649      6.91026      8.70257\n      8        -9.49354      5.56705      1.43459\n      9        -9.59955      5.60318      1.39322\n     10        -15.4077      6.37659      1.5895\n     11        -11.7703      5.83045      1.80338\n     12        -11.4169      2.53303      3.32719\n     13        -8.49429      2.67945      3.0094\n     14        -9.17395      2.74325      3.11174\n     15        -7.35265      2.86541      2.88627\n```\nwe can then obtain the maximum value found using\n```python\n>> optim_params.target\n-7.352654457092285\n```\nas well as the input parameters that yield it\n```python\n>> optim_params.params\n{'x': 2.865405, 'y': 2.8862667}\n```\n\n## Contributing\nEveryone can contribute to Bayex and we welcome pull requests as well as raised issues.\nPlease refer to this [contribution guide](CONTRIBUTING.md) on how to do it.\n\n\n## References\n1. [A Tutorial on Bayesian Optimization](https://arxiv.org/abs/1807.02811)\n2. [BayesianOptimization Library](https://github.com/fmfn/BayesianOptimization)\n3. [JAX: Autograd and XLA](https://github.com/google/jax)\n",
    'author': 'Albert Alonso',
    'author_email': 'alonfnt@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alonfnt/bayex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
